#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <linux/ppdev.h>

#include "arc.h"

#define TDI     0
#define TCK     1
#define TMS     2
#define TDO     4

#define RETRIES 50

static int pfd;

static unsigned char clockin(int tms, int tdi)
{
	unsigned char data, dataout;
	
	tms = tms ? 1 : 0;
	tdi = tdi ? 1 : 0;
	
	data = (1 << TDO) | (0 << TCK) | (tms << TMS) | (tdi << TDI);
	ioctl(pfd, PPWDATA, &data);
	
	ioctl(pfd, PPRSTATUS, &dataout);
	
	data = (1 << TDO) | (1 << TCK) | (tms << TMS) | (tdi << TDI);
	ioctl(pfd, PPWDATA, &data);
	
	dataout ^= 0x80;
	dataout >>= TDO;
	dataout &= 1;
	
	return dataout;
}


void lpt_openport(void)
{
         pfd = open("/dev/parport0", O_RDWR);
         if (pfd < 0) {
		 perror("Failed to open /dev/parport0");
		 exit(0);
	 }
         
	 if ((ioctl(pfd, PPEXCL) < 0) || (ioctl(pfd, PPCLAIM) < 0)) {
		 perror("Failed to lock /dev/parport0");
		 close(pfd);
		 exit(0);
	 }	
}

void test_reset(void)
{
	clockin(1, 0);  // Run through a handful of clock cycles with TMS high to make sure
	clockin(1, 0);  // we are in the TEST-LOGIC-RESET state.
	clockin(1, 0);
	clockin(1, 0);
	clockin(1, 0);
	clockin(0, 0);  // enter runtest-idle
}

void set_instr(int instr)
{
	int i;
	
	clockin(1, 0);  // enter select-dr-scan
	clockin(1, 0);  // enter select-ir-scan
	clockin(0, 0);  // enter capture-ir
	clockin(0, 0);  // enter shift-ir (dummy)
	for (i=0; i < 4; i++)
	{
		clockin(i == 3, (instr>>i)&1);
	}
	clockin(1, 0);  // enter update-ir
}

static unsigned int read_status(void)
{
	unsigned char val;
	set_instr(0x8);
	
	clockin(1, 0);  // enter select-dr-scan
	clockin(0, 0);  // enter capture-dr
	clockin(0, 0);  // enter shift-dr

	val = clockin(0, 0);
	if(val)
		return val;

	val |= clockin(0, 0) << 1;
	if(val)
		return val;

	val |= clockin(0, 0) << 2;
	
	return val;
}

static void write_reg(int reg, int data, int bits)
{
	int i;
	set_instr(reg);

	clockin(1, 0);  // enter select-dr-scan
	clockin(0, 0);  // enter capture-dr
	clockin(0, 0);  // enter shift-dr
	for(i = 0 ; i < bits ; i++)
	{
		clockin(i == (bits-1), ((data >> i) & 1));
	}
	
	clockin(1,0);   // enter update-dr
}

static unsigned int read_reg(int reg, int bits)
{
	int i, out = 0;
	set_instr(reg);

	clockin(1, 0);  // enter select-dr-scan
	clockin(0, 0);  // enter capture-dr
	clockin(0, 0);  // enter shift-dr
	for(i = 0 ; i < bits ; i++)
	{
		out |= clockin((i-1) == bits, 0) << i;
	}
	
	clockin(1,0);   // enter update-dr

	return out;
}

static unsigned int read_x(int reg, int read_selector)
{
	int i;

	test_reset();

	write_reg(0xa, reg, 32);
	write_reg(0x9, read_selector, 4);

	clockin(0, 0);  // enter runtest-idle
	
	for(i=0; i<RETRIES; i++) {
		int status = read_status();

		clockin(1,0);
		clockin(1,0);

		if(status == 4)
			break;
		else if(status == 2)
			printf("read failed\n");
	}
	if(i == RETRIES)
		printf("read failed - too many retries\n");

	return read_reg(0xb, 32);
}

static void write_x(int reg, int read_selector, int data)
{
	int i;

	test_reset();
	
	write_reg(0xb, data, 32);

	write_reg(0xa, reg, 32);

	write_reg(0x9, read_selector, 4);

	clockin(0, 0);  // enter runtest-idle
	
	for(i=0; i<RETRIES; i++) {
		int status = read_status();

		clockin(1,0);
		clockin(1,0);

		if(status == 4)
			break;
		else if(status == 2)
			printf("read failed\n");
	}
	if(i == RETRIES)
		printf("read failed - too many retries\n");
}

static unsigned int read_aux_reg(int reg)
{
	return read_x(reg, 0x6);
}

static void write_aux_reg(int reg, int data)
{
	write_x(reg, 0x2, data);
}

static unsigned int read_core_reg(int reg)
{
	return read_x(reg, 0x5);
}

static void write_mem(int addr, int data)
{
	write_x(addr, 0, data);
}

static unsigned int read_mem(int addr)
{
	return read_x(addr, 0x4);
}

static int is_cpu_halted(void)
{
	return read_aux_reg(REG_STATUS) & A4_HALT_VALUE;
}

static void halt_cpu(void)
{
	int i;
	for(i=0; i<40; i++) {
		write_aux_reg(REG_STATUS, A4_HALT_VALUE);
		if(is_cpu_halted())
			break;
	}
	
	if(i == 40) {
		printf("Failed to halt CPU\n");
		//return -1;
	}
}

static void start_cpu(void)
{
	write_aux_reg(REG_STATUS, (read_aux_reg(REG_STATUS) & ~(A4_HALT_VALUE)));
}

int main(int argc, char* argv[])
{
	unsigned int cpu_id;
	lpt_openport();

	if(argc < 2)
		return 0;

	if(!strcmp("-d", argv[1])) {
		int cpu_halted = is_cpu_halted();
		if(!cpu_halted)
			halt_cpu();
		
		cpu_id = read_aux_reg(REG_ID);
		
		printf("ID = 0x%x, PC = 0x%x, STATE = %s\n", cpu_id & 0xff, (read_aux_reg(REG_STATUS) & 0xffffff)<<2, cpu_halted ? "halted" : "running");
		printf("R0 = %x, R1 = %x, R2 = %x\n", read_core_reg(0), read_core_reg(1), read_core_reg(2));
		printf("FP = %x, SP = %x, BLINK = %x\n", read_core_reg(27), read_core_reg(28), read_core_reg(31));
		if(!cpu_halted)
			start_cpu();
	} else if(!strcmp("-u", argv[1])) {
		if(argc < 3)
			return 0;

		FILE *fd = fopen(argv[2], "r");
		if(fd == NULL)
			return 0;
		
		fseek(fd, 0, SEEK_END);
		long buffersize = ftell(fd);
		fseek(fd, 0, SEEK_SET);
		int *buffer = (int *)malloc(buffersize);
		if(buffer == NULL)
			return 0;

		fread(buffer, 1, buffersize, fd);
		fclose(fd);

		halt_cpu();

		// Set PC to 0x808000 (the 8 MB RAM wraps at this address)
		write_aux_reg(REG_STATUS, (read_aux_reg(REG_STATUS) & 0xff000000) | (0x808000>>2));

		// Upload the data at 0x808000
		long i;
		for(i=0; i<buffersize; i += 4) {
			write_mem(i+0x808000, buffer[i/4]);
		}
		
		free(buffer);

		start_cpu();
	} else if(!strcmp("-r", argv[1])) {
		if(argc < 3)
			return 0;

		int addr = 0;
		int bytes = 0x40;
		sscanf(argv[2], "%x", &addr);
		
		if(argc > 3)
			sscanf(argv[3], "%x", &bytes);

		int i;
		for(i=0; i<bytes; i+=4) {
			unsigned int data = read_mem(addr+i);
			printf("%08x ", data);
			if((i % 16) == 12)
				printf("\n");
		}
	} else if(!strcmp("-s", argv[1])) {
		if(argc < 5)
			return 0;

		FILE *fd = fopen(argv[2], "w+");
		if(fd == NULL)
			return 0;
		
		int addr = 0, len = 0;
		sscanf(argv[3], "%x", &addr);
		sscanf(argv[4], "%x", &len);
		
		halt_cpu();

		// Save the data
		int i;
		for(i=0; i<len; i += 4) {
			unsigned int data = read_mem(i+addr);
			fwrite(&data, 1, sizeof(data), fd);
		}
		
		fclose(fd);

		start_cpu();
	}
	
	return 0;
}
